/* qmconfig.cpp
 *
 * $Id: qmconfig.cpp,v 1.10 2000/12/18 17:57:37 amos Exp $
 *
 * Apollo sound player: http://www.apolloplayer.org
 * Copyright(C) 2000 Apollo Team.  See CREDITS file.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * The GNU General Public License is also available online at:
 *
 * http://www.gnu.org/copyleft/gpl.html
 */

#include "qmconfig.h"

#include <qdir.h>
#include <qdom.h>
#include <qfile.h>
#include <qmessagebox.h>
#include <qstring.h>
#include <qtextstream.h>

#include <iostream.h>

/*! \file qmconfig.cpp
  Implementation of QmConfig class
*/

auto_ptr<QmConfig> QmConfig::s_pInstance;

/*!
  \class QmConfig qmconfig.h
  \brief Provides configuration file reading and writing.

  Takes care of handling the persistence for Apollo regarding
  window geometry and music path. To add a new persistent variable
  you need to create a read and write function and add it to the correct
  read/write structure in both the load() and save() function.

  To help out in reading and writing data there exists several generalized
  function handling various structures. They are:
  readGeometry() and writeGeometry() for handling QRect geometries.
  readPath() and writePath() for handling QString path, these path strings are stored in CDATA sections to make sure special characters are allowed.
  nodeAsInteger() returns the integer value of a nodes first text child.
  indent() returns a string containg the correct indentation for a given level.

  Remember that only one instance of this class is allowed, this means that you
  cannot create this on your own, instead use QmConfig::instance() to obtain
  the pointer.
*/

/*!
  Creates a config object with default data.
*/

QmConfig::QmConfig()
    // Set default data
    : m_WindowGeometry( 0, 0, 400, 400 ),
    m_FileDialogGeometry( 0, 0, -1, -1 ),
    m_CurrentMusicPath( QDir::homeDirPath() ),
    m_Mpg123Path( "/usr/bin/mpg123" )
{
}

/*!
  Destroys the object
*/

QmConfig::~QmConfig()
{
}

/*!
  Returns the only instance of the QmConfig class. Uses auto_ptr for storing the object
  so it is automaticly destroyed when quitting.
*/

QmConfig *
QmConfig::instance()
{
    if ( QmConfig::s_pInstance.get() == 0 )
    {
        s_pInstance.reset( new QmConfig() );
    }
    return (QmConfig *)QmConfig::s_pInstance.get();
}


/*!
  Loads the configuration file from ~/.apollo/.  If the folder does not
  exist, it will be created.

  Reading is done by using a QDomDocument object.
*/

void
QmConfig::load()
{
	QDir dir( configPath() );

	if( ! dir.exists() )
	{
		QMessageBox::information( 0, QObject::tr("Welcome to Apollo"),
								  QObject::tr("It seems this is the first time you run Apollo.<br>"
											  "Welcome!  A '.apollo' folder has been created in<br>"
											  "your home folder in which Apollo will store its<br>"
											  "configuration."));
		dir.mkdir( configPath() );
		return;
	}

    QFile conf_file( configPath() + "/" + configFileName() );
    if ( !conf_file.open( IO_ReadOnly ) )
    {
        cout << "Failed to open config file \""
             << conf_file.name()
             << "\" in read mode" << endl;
        return;
    }
    QDomDocument doc;
    doc.setContent( &conf_file );
    QDomElement root = doc.documentElement();
    if ( root.nodeName() != "apollo" )
    {
        cout << "Root node of config file is not apollo, found \"" << root.nodeName() << "\", invalid config\n";
        return;
    }

    QDomNode child = root.firstChild();

    while ( !child.isNull() )
    {
        if ( child.isElement() )
        {
            if ( child.nodeName() == "window" )
                readWindow( child );
            else if ( child.nodeName() == "file-dialog" )
                readFileDialog( child );
            else if ( child.nodeName() == "path" )
                readPathConfig( child );
			else if ( child.nodeName() == "playingsong" )
				m_PlayingSong = readPlayingSong( child );
            else
            {
                cout << "Unknown element \"" << child.nodeName() <<  "\", ignoring\n";
            }
        }
        else
        {
            cout << "Unknown node \"" << child.nodeName() << "\", ignoring\n";
        }

        child = child.nextSibling();
    }
}

/*!
  Reads out window specific configurations, currently only geometry is read.
*/

void
QmConfig::readWindow( const QDomNode &node )
{
    QDomNode child = node.firstChild();

    while ( !child.isNull() )
    {
        if ( child.isElement() )
        {
            if ( child.nodeName() == "geometry" )
                m_WindowGeometry = readGeometry( child );
            else if ( child.nodeName() == "scroller" )
                m_ScrollPos = readPosition( child );
            else
            {
                cout << "window: Unknown element \"" << child.nodeName() << "\", ignoring\n";
            }
        }
        else
        {
            cout << "window: Unknown node \"" << child.nodeName() << "\", ignoring\n";
        }

        child = child.nextSibling();
    }
}

/*!
  Reads out file dialog specific configurations, currently only geometry is read.
*/

void
QmConfig::readFileDialog( const QDomNode &node )
{
    QDomNode child = node.firstChild();

    while ( !child.isNull() )
    {
        if ( child.isElement() )
        {
            if ( child.nodeName() == "geometry" )
                m_FileDialogGeometry = readGeometry( child );
            else
                cout << "file-dialog: Unknown element \"" << child.nodeName() << "\", ignoring\n";
        }
        else
        {
            cout << "file-dialog: Unknown node \"" << child.nodeName() << "\", ignoring\n";
        }

        child = child.nextSibling();
    }
}

/*!
  Reads out a general geometry structure, this structure consists of a position
  and size. The structure generally looks like
  <x>x-position</x>
  <y>y-position</y>
  <width>width</width>
  <height>height</height>

  This function is safe to resuse with different elements.
*/

QRect
QmConfig::readGeometry( const QDomNode &node ) const
{
    QRect rect;
    QDomNode child = node.firstChild();

    while ( !child.isNull() )
    {
        if ( child.isElement() )
        {
            if ( child.nodeName() == "x" )
                rect.setX( nodeAsInteger( child ) );
            else if ( child.nodeName() == "y" )
                rect.setY( nodeAsInteger( child ) );
            else if ( child.nodeName() == "width" )
                rect.setWidth( nodeAsInteger( child ) );
            else if ( child.nodeName() == "height" )
                rect.setHeight( nodeAsInteger( child ) );
            else
            {
                cout << "geometry: Unknown element \"" << child.nodeName() << "\", ignoring\n";
            }
        }
        else
        {
            cout << "geometry: Unknown node \"" << child.nodeName() << "\", ignoring\n";
        }

        child = child.nextSibling();
    }

    return rect;
}

/*!
  Reads out a general position structure, this structure consists of an x and y position.
  The structure generally looks like
  <x>x-position</x>
  <y>y-position</y>

  This function is safe to resuse with different elements.
*/

QPoint
QmConfig::readPosition( const QDomNode &node ) const
{
    QPoint pos;
    QDomNode child = node.firstChild();

    while ( !child.isNull() )
    {
        if ( child.isElement() )
        {
            if ( child.nodeName() == "x" )
                pos.setX( nodeAsInteger( child ) );
            else if ( child.nodeName() == "y" )
                pos.setY( nodeAsInteger( child ) );
            else
            {
                cout << "position: Unknown element \"" << child.nodeName() << "\", ignoring\n";
            }
        }
        else
        {
            cout << "position: Unknown node \"" << child.nodeName() << "\", ignoring\n";
        }

        child = child.nextSibling();
    }

    return pos;
}

/*!
  Reads the content of a node and returns the result as an integer. The first child of the
  node is examined, if it is a text node the integer value is returned otherwise a warning
  is issued and 0 is returned.

  This function is safe to resue for different elements.
*/

int
QmConfig::nodeAsInteger( const QDomNode &node ) const
{
    QDomNode data = node.firstChild();
    if ( !data.isText() )
    {
        cout << "node_as_integer: expected text child node, found \"" << data.nodeName() << "\"\n";
        return 0;
    }
    return data.nodeValue().toInt();
}

/*!
  Reads out the path specific configuration, currently only the current music path
  is read.
*/

void
QmConfig::readPathConfig( const QDomNode &node )
{
    QDomNode child = node.firstChild();

    while ( !child.isNull() )
    {
        if ( child.isElement() )
        {
            if ( child.nodeName() == "current" )
                m_CurrentMusicPath = readPath( child );
            else if ( child.nodeName() == "mpg123" )
                m_Mpg123Path = readPath( child );
            else
            {
                cout << "path: Unknown element \"" << child.nodeName() << "\", ignoring\n";
            }
        }
        else
        {
            cout << "path: Unknown node \"" << child.nodeName() << "\", ignoring\n";
        }

        child = child.nextSibling();
    }
}

/*!
  Reads out a general path structure, this structure consists of path inside a CDATA element.
  The structure generally looks like
  <![CDATA[/my/music/path/]]>

  This function is safe to resuse with different elements.
*/

QString
QmConfig::readPath( const QDomNode &node ) const
{
    QDomNode data = node.firstChild();
    if ( !data.isCDATASection() )
    {
        cout << "path: Excepted CDATA section, found \"" << data.nodeName() << "\"\n";
        return QString::null;
    }
    return data.nodeValue();
}


/*!
  Reads out the song which was played the last time Apollo was used.
*/

QString
QmConfig::readPlayingSong( const QDomNode &node ) const
{
    QDomNode data = node.firstChild();
    if ( !data.isCDATASection() )
    {
        cout << "path: Excepted CDATA section, found \"" << data.nodeName() << "\"\n";
        return QString::null;
    }
    return data.nodeValue();
}

/*!
  Saves the configurtion in the config file located in ~/.apollo/config.
  The configuration is stored as an XML file.
*/

void
QmConfig::save()
{
	QDir dir( configPath() );

	if( ! dir.exists() )
		dir.mkdir( configPath() );
	
    QFile conf_file( configPath() + "/" + configFileName() );
    if ( !conf_file.open( IO_WriteOnly ) )
    {
        cout << "Failed to open config file \""
             << conf_file.name()
             << "\" in write mode" << endl;
        return;
    }
    QTextStream conf_stream( &conf_file );
    conf_stream << "<!DOCTYPE APOLLO>\n"
                << "<apollo>\n";

    writeWindow( conf_stream, 1 );
    writeFileDialog( conf_stream, 1 );
    writePathConfig( conf_stream, 1 );
	writePlayingSong( conf_stream, 1 );

    conf_stream << "</apollo>\n";
}

/*!
  Writes out the window specific configuration, currently only the window geoemtry
  is stored.
*/

void
QmConfig::writeWindow( QTextStream &conf_stream, int level )
{
    conf_stream << indent( level ) << "<window>\n";
    writeWindowGeometry( m_WindowGeometry, conf_stream, level + 1 );
    conf_stream << indent( level + 1 ) << "<scroller>\n";
    writeWindowPosition( m_ScrollPos, conf_stream, level + 2 );
    conf_stream << indent( level + 1 ) << "</scroller>\n";
    conf_stream << indent( level ) << "</window>\n";
}

/*!
  Writes out the file dialog specific configuration, currently only the window geoemtry
  is stored.
*/

void
QmConfig::writeFileDialog( QTextStream &conf_stream, int level )
{
    conf_stream << indent( level ) << "<file-dialog>\n";
    writeWindowGeometry( m_FileDialogGeometry, conf_stream, level + 1 );
    conf_stream << indent( level ) << "</file-dialog>\n";
}

/*!
  General write function for geometry structures, it writes out the position
  and size of a QRect. This structure will look like.
  <x>x-position</x>
  <y>y-position</y>
  <width>width</width>
  <height>height</height>

  This function is safe to reuse for different geoemtries.
*/

void
QmConfig::writeWindowGeometry( const QRect &rect, QTextStream &conf_stream, int level )
{
    conf_stream << indent( level ) << "<geometry>\n";
    writeWindowPosition( rect.topLeft(), conf_stream, level + 1 );
    writeWindowSize( rect.size(), conf_stream, level + 1 );
    conf_stream << indent( level ) << "</geometry>\n";
}

/*!
  General write function for position, it writes out the x and y position of
  a QPoint object. This structure will look like.
  <x>x-position</x>
  <y>y-position</y>

  This function is safe to reuse for different positions.
*/

void
QmConfig::writeWindowPosition( const QPoint &pos, QTextStream &conf_stream, int level )
{
    conf_stream << indent( level ) << "<x>"
                << QString::number( pos.x() )
                << "</x>\n"
                << indent( level ) << "<y>"
                << QString::number( pos.y() )
                << "</y>\n";
}

/*!
  General write function for size, it writes out the width and height of
  a QSize object. This structure will look like.
  <width>width</width>
  <height>height</height>

  This function is safe to reuse for different sizes.
*/

void
QmConfig::writeWindowSize( const QSize &size, QTextStream &conf_stream, int level )
{
    conf_stream << indent( level ) << "<width>"
                << QString::number( size.width() )
                << "</width>\n"
                << indent( level ) << "<height>"
                << QString::number( size.height() )
                << "</height>\n";
}

/*!
  Writes out the path specific configuration, currently only the current music path
  is written.
*/

void
QmConfig::writePathConfig( QTextStream &conf_stream, int level )
{
    conf_stream << indent( level ) << "<path>\n";
    writePath( "current", m_CurrentMusicPath, conf_stream, level + 1 );
    writePath( "mpg123", m_Mpg123Path, conf_stream, level + 1 );
    conf_stream << indent( level ) << "</path>\n";
}

/*!
  General write function for path structures, it writes out the string representation
  of the path in a CDATA section. The structure will look like.
  <![CDATA[/my/music/path]]>
*/

void
QmConfig::writePath( const QString &name, const QString &path, QTextStream &conf_stream, int level )
{
    conf_stream << indent( level ) << "<" << name << ">"
                << "<![CDATA[" << path << "]]>"
                << "</" << name << ">\n";
}

/*!
  Writes out the song being currently played to the config, this allows
  Apollo to remember which song you were playing the last time.
*/

void
QmConfig::writePlayingSong( QTextStream &conf_stream, int level )
{
	conf_stream << indent( level ) << "<playingsong>"
				<< "<![CDATA[" << m_PlayingSong << "]]>"
				<< "</playingsong>\n";
}

/*!
  Creates a string which contains the correct indentation for a given
  level, each level adds 4 spaces to the string. Level 0 is an empty string.

  This is used for correctly indenting the output of the config, each
  time a sub function is called the level is increased by one.
*/

QString
QmConfig::indent( int level ) const
{
    QString tmp;
    for ( ; level > 0; --level )
        tmp += "    ";
    return tmp;
}


/*!
  Sets the position of the window to be stored.
*/

void
QmConfig::setWindowPosition( const QPoint &pos )
{
    m_WindowGeometry.moveTopLeft( pos );
}

/*!
  Sets the size of the window to be stored.
*/

void
QmConfig::setWindowSize( const QSize &size )
{
    m_WindowGeometry.setSize( size );
}

/*!
  Sets the position of the file dialog to be stored.
*/

void
QmConfig::setFileDialogPosition( const QPoint &pos )
{
    m_FileDialogGeometry.moveTopLeft( pos );
}

/*!
  Sets the size of the file dialog to be stored.
*/

void
QmConfig::setFileDialogSize( const QSize &size )
{
    m_FileDialogGeometry.setSize( size );
}

void
QmConfig::setScrollPosition( const QPoint &pos )
{
    m_ScrollPos = pos;
}

/*!
  Sets the current music path to be stored.
*/

void
QmConfig::setCurrentMusicPath( const QString &path )
{
    m_CurrentMusicPath = path;
}

/*!
  Sets the current song being played to be stored.
*/

void
QmConfig::setPlayingSong( const QString &song )
{
	m_PlayingSong = song;
}

/*!
  Sets the path to the mpg123 program, default is /usr/bin/mpg123
*/

void
QmConfig::setMpg123Path( const QString &path )
{
    m_Mpg123Path = path;
}

/*!
  Returns the stored window position.
*/

QPoint
QmConfig::windowPosition() const
{
    return m_WindowGeometry.topLeft();
}

/*!
  Returns the stored window size.
*/

QSize
QmConfig::windowSize() const
{
    return m_WindowGeometry.size();
}

/*!
  Returns the stored file dialog position.
*/

QPoint
QmConfig::fileDialogPosition() const
{
    return m_FileDialogGeometry.topLeft();
}

/*!
  Returns the stored file dialog size.
*/

QSize
QmConfig::fileDialogSize() const
{
    return m_FileDialogGeometry.size();
}

const QPoint &
QmConfig::scrollPosition() const
{
    return m_ScrollPos;
}

/*!
  Returns the stored music path.
*/

const QString &
QmConfig::currentMusicPath() const
{
    return m_CurrentMusicPath;
}

/*!
  Returns the stored playing song.
*/

const QString &
QmConfig::playingSong() const
{
	return m_PlayingSong;
}

/*!
  Returns the path to the mpg123 program, default is /usr/bin/mpg123
*/

const QString &
QmConfig::mpg123Path() const
{
    return m_Mpg123Path;
}

/*!
  Returns the path of the config file. This is not including the filename.
*/

QString
QmConfig::configPath() const
{
    return QString( QDir::homeDirPath() + "/.apollo" );
}

/*!
  Returns the file name of the config file. This is not including the file path.
*/

QString
QmConfig::configFileName() const
{
    return QString( "config" );
}
