/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*  This file is part of MixMagic.

    Copyright (C) 2000 Jonas Borgstrm <jonas_b@bitsmart.com>.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public License
    along with this library; see the file COPYING.LIB.  If not, write to
    the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
    Boston, MA 02111-1307, USA.
*/
#include "mixmagic.h"
#include "loopblock.h"
#include "track.h"
#include "mixer.h"

static BlockClass *parent_class = NULL;
static void loopblock_class_init (LoopBlockClass *class);
static void loopblock_init (LoopBlock *lblock);
static gboolean loopblock_configure (GtkWidget *widget,GdkEventConfigure *event);
static void mix_into_buffer (Block *block, gint16 *buffer, gint position, gint length);
static void lower_left_box_motion (Block *block, GdkEventMotion *event, gint x, gint y);
static void lower_right_box_motion (Block *block, GdkEventMotion *event, gint x, gint y);

guint
loopblock_get_type ()
{
        static guint loopblock_type = 0;
	
        if (!loopblock_type) {

		GtkTypeInfo loopblock_info = {
			"LoopBlock",
			sizeof (LoopBlock),
			sizeof (LoopBlockClass),
			(GtkClassInitFunc) loopblock_class_init,
			(GtkObjectInitFunc) loopblock_init,
			(GtkArgSetFunc) NULL,
			(GtkArgGetFunc) NULL
		};
	    
		loopblock_type = gtk_type_unique (block_get_type (), &loopblock_info);
          }
	
        return loopblock_type;
}

static void
loopblock_class_init (LoopBlockClass *class)
{
	GtkWidgetClass *widget_class;
	GtkObjectClass *object_class;
	BlockClass *block_class;

	parent_class = gtk_type_class (block_get_type ());

	object_class = (GtkObjectClass*) class;
	block_class = (BlockClass*) class;

	block_class->mix_into_buffer = mix_into_buffer;
	block_class->lower_left_box_motion = lower_left_box_motion;
	block_class->lower_right_box_motion = lower_right_box_motion;
	block_class->upper_half_motion = parent_class->lower_half_motion;

	widget_class = (GtkWidgetClass*) class;
	widget_class->configure_event = loopblock_configure;
}

static void
loopblock_init (LoopBlock *lblock)
{
}

static void
lower_left_box_motion (Block *block, GdkEventMotion *event, gint x, gint y)
{
	LoopBlock *lblock = LOOPBLOCK (block);
	gint change;
	
	if (!block->click_flag)
		return;

	change = block->click_point.x - x;
	/* Only valid values */
	
	/* Don't make the block to short */
	if (block->length + change * block->track->song->x_scale < BLOCK_MIN_WIDTH)
		change = (BLOCK_MIN_WIDTH - block->length) / block->track->song->x_scale;
	
	if (change * block->track->song->x_scale > block->position)
		change = block->position / block->track->song->x_scale;
	
	if (-change * block->track->song->x_scale > block->length)
		change = block->length / block->track->song->x_scale;
	
	block->position -= block->track->song->x_scale * change;
	block->start -= block->track->song->x_scale * change;
	while (block->start < 0)
		block->start += lblock->loop_length;
	while (block->start >= lblock->loop_length)
		block->start -= lblock->loop_length;
	
	block->length += block->track->song->x_scale * change;

	/* Update the widget */
	block_update (block);
}

static void
lower_right_box_motion (Block *block, GdkEventMotion *event, gint x, gint y)
{
	gint change;
	
	if (!block->click_flag)
		return;

	change = x - block->click_point.x;
	
				/* Don't make the block to short */
	if (block->length + change * block->track->song->x_scale < BLOCK_MIN_WIDTH)
		change = (BLOCK_MIN_WIDTH - block->length) / block->track->song->x_scale;
	
	block->length += block->track->song->x_scale * change;
	block->click_point.x += change;

	/* Update the widget */
	block_update (block);
}

static gboolean 
loopblock_configure (GtkWidget *widget,GdkEventConfigure *event)
{
	gint i, width, height, value, offset;
	Block *block = BLOCK (widget);
	LoopBlock *lblock = LOOPBLOCK (widget);

	width = widget->allocation.width;
	height = widget->allocation.height;

	/* Only update the pixmap when needed */
	if (block->pixmap && 
	    block->old_width == width &&
	    block->old_height == height) {

		return TRUE;
	}

	block->old_width = width;
	block->old_height = height;

	if (block->wav_gc == NULL)
		return TRUE;

	if (block->pixmap)
		gdk_pixmap_unref (block->pixmap);

	block->pixmap = gdk_pixmap_new (widget->window, width, height, -1);	
	
	gdk_draw_rectangle (block->pixmap,
			    widget->style->white_gc,
			    TRUE,
			    0,
			    0,
			    width,
			    height);

	offset = (block->start + lblock->loop_start)/ 8;

	for (i = 0; i < width;i++) {

#if 1
		value = index_get_average (block->sample->index, 
					   offset, lblock->loop_start / 8,
					   block->track->song->x_scale / 8, 
					   lblock->loop_length / 8) / (0x7f / height);
		gdk_draw_line (block->pixmap,
			       block->wav_gc,
			       i, height / 2 + value,
			       i, height / 2 - value);
#else
		value = index_get_average (block->sample->index, 
					   offset, lblock->loop_start / 8,
					   block->track->song->x_scale / 8, 
					   lblock->loop_length / 8) / (0x7f / height);

		gdk_draw_line (block->pixmap,
			       block->wav_gc,
			       i, height - value,
			       i, height);
#endif
		offset += block->track->song->x_scale / 8;
#if 0
		if ((gint)data >= (gint)block->sample->index->data + (lblock->loop_length / 8))
			data -= (lblock->loop_length / 8);
#endif
	}
	
	gdk_draw_rectangle (block->pixmap,
			    widget->style->fg_gc[widget->state],
			    FALSE,
			    0,
			    0,
			    width - 1,
			    height - 1);

	/* Draw a black "handle" in the lower left corner */
        gdk_draw_rectangle (block->pixmap,
                            widget->style->black_gc,
                            TRUE,
                            0,
                            height - BLOCK_HANDLE_SIZE,
                            BLOCK_HANDLE_SIZE,
                            BLOCK_HANDLE_SIZE);

	/* Draw a black "handle" in the lower right corner */
        gdk_draw_rectangle (block->pixmap,
                            widget->style->black_gc,
                            TRUE,
                            width - BLOCK_HANDLE_SIZE,
                            height - BLOCK_HANDLE_SIZE,
                            BLOCK_HANDLE_SIZE,
                            BLOCK_HANDLE_SIZE);

	gdk_draw_text (block->pixmap,
		       widget->style->font,
		       widget->style->fg_gc[widget->state],
		       0, 10,
		       block->sample->id,
		       strlen (block->sample->id));

	return TRUE;
}

GtkWidget*
loopblock_new (Track *track)
{
	LoopBlock *lblock;

        lblock = LOOPBLOCK (gtk_type_new (loopblock_get_type ()));

	lblock->block.track = track;

	return GTK_WIDGET (lblock);
}

GtkWidget*
loopblock_new_from_block (Block *block, Track *track)
{
	LoopBlock *lblock;

        lblock = LOOPBLOCK (gtk_type_new (loopblock_get_type ()));

	lblock->block.track = track;
	lblock->loop_start = block->start;
	lblock->block.start = 0;
	lblock->block.length = block->length;
	lblock->loop_length = block->length;
	lblock->block.position = block->position;
	lblock->block.sample = sample_ref (block->sample);

	return GTK_WIDGET (lblock);
}

static void
mix_into_buffer (Block *block, gint16 *buffer, gint position, gint length)
{
	LoopBlock *lblock = LOOPBLOCK (block);
	gint16 *data;
	gint i;

	data = (gint16* )(block->sample->data + lblock->loop_start * BYTES_PER_SAMPLE +
		(((block->start + position - block->position)%lblock->loop_length)* BYTES_PER_SAMPLE));

	for (i = 0;i < length / sizeof (guint16);i+=2) {

		if ((guint)data >= (guint)block->sample->data +
		    ((lblock->loop_start + lblock->loop_length) * BYTES_PER_SAMPLE)) {
			data -= (lblock->loop_length * BYTES_PER_SAMPLE) / 2;
		}

		mixer_mix_sample_into_buffer (buffer,
					      data,
					      block->amp_env, 
					      block->pan_env, 
					      block->track->amp,
					      block->track->pan,
					      position - block->position + i);

		data += 2;
		buffer += 2;
	}
}

void
loopblock_load (Song *song, Track *track, xmlNodePtr node, xmlNsPtr ns, gint track_num) {
	Block *block;
	LoopBlock *lblock;
	gchar *ptr;
	xmlNodePtr cur;
	Envelope *envelope;

	lblock = LOOPBLOCK (loopblock_new (track));
	block = BLOCK (lblock);
	
	block->start = atoi ((ptr = xmlGetProp (node, "start")));
	g_free (ptr);
	block->length = atoi ((ptr = xmlGetProp (node, "length")));
	g_free (ptr);
	block->position = atoi ((ptr = xmlGetProp (node, "position")));
	g_free (ptr);
	lblock->loop_start = atoi ((ptr = xmlGetProp (node, "loop_start")));
	g_free (ptr);
	lblock->loop_length = atoi ((ptr = xmlGetProp (node, "loop_length")));
	g_free (ptr);
	ptr = xmlGetProp (node, "sample");
	block->sample = sample_ref (g_hash_table_lookup (song->sample_hash, ptr));
	g_free (ptr);
	
	/* Add the block to out list */
	song_insert_block (song, block, track_num);
	
	cur = node->childs;
	
	while (cur) {
		
		if ((!strcmp(cur->name, "ampenv")) && (cur->ns == ns)) {
			envelope = g_new0 (Envelope, 1);
			
			envelope->x = atoi ((ptr = xmlGetProp (cur, "x")));
			g_free (ptr);
			envelope->y = atof ((ptr = xmlGetProp (cur, "y")));
			g_free (ptr);
			block->amp_env = g_list_append (block->amp_env, envelope);
		}
		
		if ((!strcmp(cur->name, "panenv")) && (cur->ns == ns)) {
			envelope = g_new0 (Envelope, 1);
			
			envelope->x = atoi ((ptr = xmlGetProp (cur, "x")));
			g_free (ptr);
			envelope->y = atof ((ptr = xmlGetProp (cur, "y")));
			g_free (ptr);
			block->pan_env = g_list_append (block->pan_env, envelope);
		}
		cur = cur->next;
	};
}
