/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*  This file is part of MixMagic.

    Copyright (C) 2000 Jonas Borgstrm <jonas_b@bitsmart.com>.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public License
    along with this library; see the file COPYING.LIB.  If not, write to
    the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
    Boston, MA 02111-1307, USA.
*/
#include <pthread.h>

#include "mixmagic.h"
#include "application.h"
#include "block.h"
#include "track.h"
#include "mixer.h"
#include "song.h"
#include "plugins.h"

static pthread_t mix_thread;
static gboolean thread_running = FALSE;

/*
 * Returns active blocks for a specific song position.
 */
static gint
mixer_get_blocks (Song *song, Block **blocks, gint position, guint *num_blocks)
{
	GList *track_list, *block_list;
	gint length = G_MAXINT;
	Track *track;
	Block *block;
	gint track_num, prev_block;

	g_assert (song != NULL);
	g_assert (blocks != NULL);

	if (position < 0)
		return -1;

	*num_blocks = 0;

	track_list = song->track_list;

	track_num = 0;
	prev_block = -1;

	while (track_list) {

		track = track_list->data;
		block_list = track->block_list;
		
		while (block_list) {
			block = block_list->data;

			if (block->position <= position && position < (block->position + block->length)) {
				/* Check if this block is on the same track as the last one */
				if (prev_block == track_num)
					(*num_blocks)--;
				else
					prev_block = track_num;

				if (track->mute_flag == FALSE) {
					blocks[*num_blocks] = block;
					(*num_blocks)++;
				}
				if (block->position + block->length - position < length)
					length = block->position + block->length - position;
 			}
			if (block->position > position) {
				if (block->position - position < length)
					length = block->position - position;

				break;
			}
			block_list = g_list_next (block_list);
		}
		track_num++;
		track_list = g_list_next (track_list);
	}
	if (length == G_MAXINT)
		length = -1;

	return length;
}

/*
 * Returns the scale value.
 */
float
mixer_calc_envelope (GList *list, gint pos)
{
	Envelope *e1;
	Envelope *e2 = NULL;
	float result;
		
	while (g_list_next (list)) {
		e2 = list->next->data;
		if (e2->x >= pos)
			break;
		list = g_list_next (list);
	}

	e1 = list->data;
	if (g_list_next (list))
		e2 = g_list_next (list)->data;
	
	if (e1->x <= pos && g_list_next (list)) {
		result = e1->y + (e2->y - e1->y )*
			((float) (pos - e1->x) /
			 ( e2->x - e1->x) );
	} else if (e1->x >= pos) {
		result = 0.5 + (e1->y - 0.5)*
			  ((float) (pos) /
			   ( e1->x));
	} else
		result = (float) e1->y;

	return result;	
}

/*
 * Returns active blocks for a specific song position.
 */
static gint
mixer_mix_blocks (Song *song, Block **blocks, gint position, gint num_blocks, gint len)
{
	gint16 buffer[MIX_BUFFER_SIZE / sizeof (gint16)];
	gint i, length;

	g_assert (song != NULL);
	g_assert (blocks != NULL);

	/* Do we need to stop? */
	if (!thread_running)
		return -1;

	length = len * BYTES_PER_SAMPLE > MIX_BUFFER_SIZE ? MIX_BUFFER_SIZE : len * BYTES_PER_SAMPLE;

	/* Clear the buffer */
	memset (buffer, 0, length);
	/* Mix the blocks */
	for (i=0;i < num_blocks;i++) {

		block_mix_into_buffer (blocks[i], buffer, position, length);

	}
	prefs->output_plugin->write(buffer, length);

	return position + length / BYTES_PER_SAMPLE;
}

/*
 * This function return FALSE, if a output_plugin is missing, else it returns TRUE
 */

static gboolean
mixer_check_output_plugin ()
{
	if (prefs->output_plugin == NULL) {

		gtk_widget_show (gnome_message_box_new ("You need to select an output plugin to be able to get sound output.",
                                                        GNOME_MESSAGE_BOX_ERROR,
                                                        GNOME_STOCK_BUTTON_OK,
                                                        NULL));
		return FALSE;
	}
	return TRUE;
}

/*
 * Plays a song.
 */
void
mixer_play_song (Song *song)
{	
	gint num_blocks; 
	gint length, old_position = 0;
	Block *blocks[MAX_TRACKS];

	g_assert (song != NULL);

	for(;;) {
		length = mixer_get_blocks (song, blocks, song->marker_position, &num_blocks);
		/* Are we at the end of the song? */
		if (length == -1) {
			if (song->repeat_flag == TRUE) {

				song->marker_position = 0;
				continue;
			} else
				break;
		}		
		old_position = song->marker_position;
		song->marker_position = mixer_mix_blocks (song, blocks, song->marker_position, num_blocks, length);
		/* Should we stop? */
		if (song->marker_position == -1)
			break;
	}
	song->marker_position = old_position;
}

/*
 * The mix thread entry point.
 */
static void *
mix_thread_start (void *arg)
{
	g_assert (arg != NULL);

	if (prefs->output_plugin->open (AFORMAT_S16_LE, SAMPLE_RATE, 2) == FALSE)
		return NULL;

	mixer_play_song ((Song *)arg);
	thread_running = FALSE;

	prefs->output_plugin->close ();

	return NULL;
}

/*
 * The play block thread entry point.
 */
static void *
play_block_thread_start (void *arg)
{
	PlayBlockArgs *args = (PlayBlockArgs *)arg;

	g_assert (arg != NULL);

	if (prefs->output_plugin->open (AFORMAT_S16_LE, SAMPLE_RATE, 2) == FALSE)
		return NULL;

	while (args->song->marker_position < args->block->length + args->block->position &&
	       args->song->marker_position != -1) { /* -1 = Die thread die! */

		args->song->marker_position = mixer_mix_blocks (args->song, 
								&args->block, 
								args->song->marker_position, 
								1, 
								args->block->length + 
								args->block->position - 
								args->song->marker_position);

		/* Should we loop? */
		if (args->song->marker_position >= args->block->length + args->block->position &&
		    args->song->repeat_flag)
			args->song->marker_position = args->block->position;
	}

	/* delete block */
	gtk_object_unref (GTK_OBJECT (args->block));
	thread_running = FALSE;

	prefs->output_plugin->close ();

	g_free (args);

	return NULL;
}

/*
 * Returns true if the thread is running.
 */
gboolean
mixer_is_playing ()
{
	return thread_running;
}

/*
 * Stop mixing thread.
 */
void 
mixer_stop_thread ()
{
	if (thread_running == TRUE) {

		thread_running = FALSE;
		pthread_join (mix_thread, NULL);
	}
}

/*
 * Plays a song using a thread.
 */
void 
mixer_play_song_threaded (Song *song)
{
	g_assert (song != NULL);

	if (mixer_check_output_plugin () == FALSE)
		return;

	/* First stop the thread if it's already running */
	mixer_stop_thread ();

	thread_running = TRUE;
	/* Start the mix thread */
	pthread_create(&mix_thread, NULL, mix_thread_start, song);
}

/*
 * Plays a block and then deletes it.
 */
void 
mixer_play_block_and_delete (Song *song, Block *block)
{
	PlayBlockArgs *args;

	g_assert (song != NULL);
	g_assert (block != NULL);

	if (mixer_check_output_plugin () == FALSE)
		return;

	/* Send the arguments in a structure */
	args = g_new0 (PlayBlockArgs, 1);
	args->song = song;
	args->block = block;

	/* First stop the thread if it's already running */
	mixer_stop_thread ();

	thread_running = TRUE;

	/* Start the play_block_thread */
	pthread_create(&mix_thread, NULL, play_block_thread_start, args);
}

void
mixer_mix_sample_into_buffer (gint16 *buffer, gint16 *data, GList *amp_env, GList *pan_env, gfloat t_amp, gfloat t_pan, gint position)
{
	gint i;
	gint16 tmp;
	double amp_scale, pan_pos, pan_left = 1, pan_right = 1, t_pan_left = 1, t_pan_right = 1;

	/* Use amplitude envelope? */
	if (amp_env)
		amp_scale = mixer_calc_envelope (amp_env, 
						 position);
	else 
		amp_scale = 1;
	
	/* Use pan envelope? */
	if (pan_env)
		pan_pos = mixer_calc_envelope (pan_env, 
					       position);
	else 
		pan_pos = 0;

	/* Block paning */
	if (pan_pos > 0) {
		pan_right = 1.0;
		pan_left = 1.0 - pan_pos;
	} else {
		pan_left = 1.0;
		pan_right = 1.0 + pan_pos;
	}
	/* Track paning */
	if (t_pan > 0) {
		t_pan_right = 1.0;
		t_pan_left = 1.0 - t_pan;
	} else {
		t_pan_left = 1.0;
		t_pan_right = 1.0 + t_pan;
	}

	/* One time for the left channel and one time for the right channel */
	for (i = 0;i < 2; i++) {
		tmp = *data;
		
		/* Left? */
		if (i == 0)
			tmp = t_amp * tmp * amp_scale * pan_left * t_pan_left;
		else
			tmp = t_amp * tmp * amp_scale * pan_right * t_pan_right;
		
		/* Check for over/underflow */
		if ((gint32)tmp + (gint32)*buffer > 0x7fff)
			*buffer= 0x7fff;
		else if ((gint32)tmp + (gint32)*buffer < -0x7fff)
			*buffer = -0x7fff;
		else
			*buffer += tmp;
		buffer++;
		data++;
	}
}
