/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*  This file is part of MixMagic.

    Copyright (C) 2000 Jonas Borgstrm <jonas_b@bitsmart.com>.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public License
    along with this library; see the file COPYING.LIB.  If not, write to
    the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
    Boston, MA 02111-1307, USA.
*/
#include <gnome.h>
#include "gmodule.h"
#include "mixmagic.h"
#include "application.h"
#include "preferences.h"

static GtkWidget *pref_win = NULL;
static GtkWidget *output_menu = NULL;
static GtkWidget *oconfigure_button = NULL;
static GtkWidget *oabout_button = NULL;

/*
 * This function is used to find an output plugin in the list by the info text (user_data)
 */

static gint
match_output_plugin_info (gconstpointer data, gconstpointer user_data)
{
	return strcmp (((OutputPlugin *)data)->info, user_data);
}

void
preferences_load (Preferences *pref)
{
	gchar *output_info;
	GList *list;
	output_info = gnome_config_get_string ("/mixmagic/plug-ins/output=MixMagic OSS output plug-in.");

	/* Try to find a mathing plugin in the list */
	list = g_list_find_custom (plugins_get_output_list (), output_info, match_output_plugin_info);
	g_free (output_info);

	if (list)
		pref->output_plugin = (OutputPlugin *)list->data;
	else if (plugins_get_output_list ()) /* Use the first plugin in the list as a default value */
		pref->output_plugin = (OutputPlugin *)plugins_get_output_list ()->data;
	else
		pref->output_plugin = NULL;
}

void
preferences_save (Preferences *pref)
{
	if (pref->output_plugin)
		gnome_config_set_string("/mixmagic/plug-ins/output", pref->output_plugin->info);

	gnome_config_sync ();
}

static void 
cancel_cb (void)
{
	pref_win = NULL;
}

static void
update_output_buttons (void)
{
	OutputPlugin *plugin;

	gtk_widget_set_sensitive (oconfigure_button, FALSE);
	gtk_widget_set_sensitive (oabout_button, FALSE);

	plugin = gtk_object_get_data (GTK_OBJECT (GTK_OPTION_MENU (output_menu)->menu_item), "plugin");

	if (plugin == NULL)
		return;

	if (plugin->configure)
		gtk_widget_set_sensitive (oconfigure_button, TRUE);
		
	if (plugin->about)
		gtk_widget_set_sensitive (oabout_button, TRUE);
}

void
output_changed_cb (GtkWidget *widget, gpointer data)
{
	update_output_buttons ();
	gnome_property_box_changed (GNOME_PROPERTY_BOX(pref_win));
}

static void 
apply_cb (void)
{
	prefs->output_plugin = gtk_object_get_data (GTK_OBJECT (GTK_OPTION_MENU (output_menu)->menu_item), "plugin");

	preferences_save (prefs);
}

static void 
output_configure_cb (void)
{
	OutputPlugin *plugin;

	plugin = gtk_object_get_data (GTK_OBJECT (GTK_OPTION_MENU (output_menu)->menu_item), "plugin");

	if (plugin == NULL)
		return;

	if (plugin->configure)
		plugin->configure ();
}

static void 
output_about_cb (void)
{
	OutputPlugin *plugin;

	plugin = gtk_object_get_data (GTK_OBJECT (GTK_OPTION_MENU (output_menu)->menu_item), "plugin");

	if (plugin == NULL)
		return;

	if (plugin->about)
		plugin->about ();
}

static void
fill_output_option_menu (GtkOptionMenu *option_menu)
{
	GtkWidget *menu, *item;
	GList *list;
	gchar *info;
	OutputPlugin *plugin;
	gint index = 0, history = 0;

	menu = gtk_menu_new ();

	list = plugins_get_output_list ();

	while (list) {

		plugin = (OutputPlugin *)list->data;

		if (prefs->output_plugin == plugin)
			history = index;

		info = g_strdup_printf ("%s (%s)", 
					plugin->info, 
					g_basename (g_module_name (plugin->module)));

		item = gtk_menu_item_new_with_label(info);
		g_free (info);
		gtk_widget_show(item);

		gtk_signal_connect(GTK_OBJECT(item), "activate",
				   GTK_SIGNAL_FUNC(output_changed_cb), NULL);

		gtk_object_set_data (GTK_OBJECT (item), "plugin", plugin);
                gtk_menu_append(GTK_MENU(menu), item);

		list = g_list_next (list);
		index++;
	}
	gtk_option_menu_remove_menu (GTK_OPTION_MENU (option_menu));
        gtk_option_menu_set_menu (GTK_OPTION_MENU(option_menu), menu);
        gtk_option_menu_set_history (GTK_OPTION_MENU(option_menu), history);
}

static GtkWidget *
create_page1 ()
{
	GtkWidget *frame;
	GtkWidget *vbox;
	GtkWidget *bbox;

	frame = gtk_frame_new("Output plug-in");
	gtk_container_border_width(GTK_CONTAINER(frame), GNOME_PAD_SMALL);

	vbox = gtk_vbox_new (FALSE, GNOME_PAD_SMALL);
	gtk_container_border_width(GTK_CONTAINER(vbox), GNOME_PAD_SMALL);

	bbox = gtk_hbutton_box_new ();
	gtk_button_box_set_layout(GTK_BUTTON_BOX(bbox), GTK_BUTTONBOX_START);
	gtk_button_box_set_spacing(GTK_BUTTON_BOX(bbox), 10);

	output_menu = gtk_option_menu_new ();
	fill_output_option_menu (GTK_OPTION_MENU (output_menu));

	gtk_box_pack_start (GTK_BOX (vbox), output_menu, TRUE, FALSE, 0);
	gtk_box_pack_start (GTK_BOX (vbox), bbox, TRUE, FALSE, 0);

	oconfigure_button = gtk_button_new_with_label ("Configure");
	gtk_box_pack_start (GTK_BOX (bbox), oconfigure_button, TRUE, FALSE, 0);
	gtk_signal_connect(GTK_OBJECT(oconfigure_button), "clicked",
			   GTK_SIGNAL_FUNC(output_configure_cb), NULL);


	oabout_button = gtk_button_new_with_label ("About");
	gtk_box_pack_start (GTK_BOX (bbox), oabout_button, TRUE, FALSE, 0);
	gtk_signal_connect(GTK_OBJECT(oabout_button), "clicked",
			   GTK_SIGNAL_FUNC(output_about_cb), NULL);

	gtk_container_add(GTK_CONTAINER(frame), vbox);

	update_output_buttons ();

	return frame;
}

/*
 * This function vill display a preferences dialog, 
 * if you call it when the dialog is alreay showing, it will be hidden
 */

void
preferences_show ()
{
	GtkWidget *label;
	if (pref_win) {

		gtk_widget_destroy (pref_win);
		cancel_cb ();

		return;
	}
	pref_win = gnome_property_box_new ();
	gtk_widget_realize (pref_win);

	label = gtk_label_new("Preferences");
	gtk_notebook_append_page(GTK_NOTEBOOK(GNOME_PROPERTY_BOX(pref_win)->notebook),
				 create_page1 (), label);

	gtk_signal_connect(GTK_OBJECT(pref_win), "apply",
			   GTK_SIGNAL_FUNC(apply_cb), NULL);
	gtk_signal_connect(GTK_OBJECT(pref_win), "destroy",
			   GTK_SIGNAL_FUNC(cancel_cb), NULL);
	gtk_signal_connect(GTK_OBJECT(pref_win), "delete_event",
			   GTK_SIGNAL_FUNC(cancel_cb), NULL);

	gtk_window_set_modal (GTK_WINDOW (pref_win), TRUE);
	gtk_widget_show_all (pref_win);
}
