/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*  This file is part of MixMagic.

    Copyright (C) 2000 Jonas Borgstrm <jonas_b@bitsmart.com>.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public License
    along with this library; see the file COPYING.LIB.  If not, write to
    the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
    Boston, MA 02111-1307, USA.
*/

#include <gnome.h>
#include <tree.h>
#include <parser.h>

#include "mixmagic.h"
#include "song.h"
#include "sample.h"
#include "mixer.h"
#include "block.h"
#include "loopblock.h"
#include "track.h"

static GtkVBoxClass *parent_class = NULL;

static void song_finalize (GtkObject *object);
static void song_class_init (SongClass *class);
static void song_init (Song *song);
static void save_loopblock (LoopBlock *lblock, xmlNodePtr tree);

guint
song_get_type ()
{
        static guint song_type = 0;
	
        if (!song_type) {

		GtkTypeInfo song_info = {
			"Song",
			sizeof (Song),
			sizeof (SongClass),
			(GtkClassInitFunc) song_class_init,
			(GtkObjectInitFunc) song_init,
			(GtkArgSetFunc) NULL,
			(GtkArgGetFunc) NULL
		};
	    
		song_type = gtk_type_unique (gtk_vbox_get_type (), &song_info);
          }
	
        return song_type;
}

static void
song_class_init (SongClass *class)
{
	GtkWidgetClass *widget_class;
	GtkObjectClass *object_class;

	parent_class = gtk_type_class (GTK_TYPE_VBOX);

	object_class = (GtkObjectClass*) class;
	object_class->finalize = song_finalize;

	widget_class = (GtkWidgetClass*) class;
}

static void
song_init (Song *song)
{
	song->scrollbar = gtk_hscrollbar_new (NULL);

	gtk_box_pack_end (GTK_BOX (song), song->scrollbar, FALSE, TRUE, 0);
	gtk_widget_show (song->scrollbar);

	song->sample_hash = g_hash_table_new (g_str_hash, g_str_equal);

	song->x_scale = 10000;
	song->block_height = 100;

	song->marker_id = -1;
}

static void
unref_sample (gpointer key, gpointer value, gpointer user_data)
{
	sample_unref ((Sample *)value);
}

static void 
song_finalize (GtkObject *object)
{
	Song *song = SONG (object);

	g_print ("Song finalize\n");

	g_list_free (song->track_list);

	/* unref samples */
	g_hash_table_foreach (song->sample_hash, unref_sample, NULL);
	g_hash_table_destroy (song->sample_hash);

	(*GTK_OBJECT_CLASS (parent_class)->finalize) (object);
}

/*
 * Allocates a new mix file.
 */
GtkWidget *
song_new (GnomeAppBar *appbar)
{
	Song *song;
	Track *track;

	song = SONG (gtk_type_new (song_get_type ()));
	song->appbar = appbar;

	/* Insert two empty tracks */
	track = TRACK (track_new (song_get_adjustment (song), song, 0));
	gtk_widget_show (GTK_WIDGET (track));
	gtk_box_pack_start (GTK_BOX (song),
			    GTK_WIDGET (track), FALSE, FALSE, 0);
	
	song->track_list = g_list_append (song->track_list, track);

	return GTK_WIDGET (song);
}

GtkAdjustment*
song_get_adjustment (Song *song)
{
        g_return_val_if_fail (song != NULL, NULL);
        g_return_val_if_fail (IS_SONG (song), NULL);
	
        return gtk_range_get_adjustment (GTK_RANGE (song->scrollbar));
}

void
song_set_adjustment (Song *song,
		      GtkAdjustment *adjustment)
{
        g_return_if_fail (song != NULL);
        g_return_if_fail (IS_SONG (song));
	
        gtk_range_set_adjustment (GTK_RANGE (song->scrollbar), adjustment);
}

/*
 * Returns TRUE if the song is empty.
 */
gboolean
song_is_empty (Song *song)
{
	if (g_hash_table_size (song->sample_hash))
		return FALSE;

	if (g_list_length (song->track_list) > 1)
		return FALSE;

	return TRUE;
}

static void
save_sample (gpointer key, Sample *sample, xmlNodePtr tree)
{
	xmlNodePtr subtree;

	subtree = xmlNewChild(tree, NULL, "sample", NULL);
	xmlSetProp(subtree, "id", sample->id);
	xmlSetProp(subtree, "filename", sample->filename);
}

static void
save_amp_env (Envelope *env, xmlNodePtr tree)
{
	xmlNodePtr subtree;
	gchar buffer[80];

	subtree = xmlNewChild(tree, NULL, "ampenv", NULL);
	sprintf (buffer, "%d", env->x);
	xmlSetProp(subtree, "x", buffer);
	sprintf (buffer, "%f", env->y);
	xmlSetProp(subtree, "y", buffer);
}

static void
save_pan_env (Envelope *env, xmlNodePtr tree)
{
	xmlNodePtr subtree;
	gchar buffer[80];

	subtree = xmlNewChild(tree, NULL, "panenv", NULL);
	sprintf (buffer, "%d", env->x);
	xmlSetProp(subtree, "x", buffer);
	sprintf (buffer, "%f", env->y);
	xmlSetProp(subtree, "y", buffer);
}

static void
save_block (Block *block, xmlNodePtr tree)
{
	xmlNodePtr subtree;
	gchar buffer[80];

	if (IS_LOOPBLOCK (block)) {

		save_loopblock (LOOPBLOCK (block), tree);
		return;
	}

	subtree = xmlNewChild(tree, NULL, "block", NULL);
	xmlSetProp(subtree, "sample", block->sample->id);
	sprintf (buffer, "%d", block->start);
	xmlSetProp(subtree, "start", buffer);
	sprintf (buffer, "%d", block->length);
	xmlSetProp(subtree, "length", buffer);
	sprintf (buffer, "%d", block->position);
	xmlSetProp(subtree, "position", buffer);

	g_list_foreach (block->amp_env, (GFunc)save_amp_env, subtree);
	g_list_foreach (block->pan_env, (GFunc)save_pan_env, subtree);
}

static void
save_loopblock (LoopBlock *lblock, xmlNodePtr tree)
{
	xmlNodePtr subtree;
	gchar buffer[80];
	Block *block = BLOCK (lblock);

	subtree = xmlNewChild(tree, NULL, "loopblock", NULL);
	xmlSetProp(subtree, "sample", block->sample->id);
	sprintf (buffer, "%d", block->start);
	xmlSetProp(subtree, "start", buffer);
	sprintf (buffer, "%d", block->length);
	xmlSetProp(subtree, "length", buffer);
	sprintf (buffer, "%d", block->position);
	xmlSetProp(subtree, "position", buffer);
	/* LoopBlock Specific */
	sprintf (buffer, "%d", lblock->loop_start);
	xmlSetProp(subtree, "loop_start", buffer);
	sprintf (buffer, "%d", lblock->loop_length);
	xmlSetProp(subtree, "loop_length", buffer);

	g_list_foreach (block->amp_env, (GFunc)save_amp_env, subtree);
	g_list_foreach (block->pan_env, (GFunc)save_pan_env, subtree);
}

static void
save_track (Track *track, xmlNodePtr tree)
{
	xmlNodePtr subtree;
	gchar buffer[80];

	subtree = xmlNewChild(tree, NULL, "track", NULL);
	sprintf (buffer, "%f", track->amp);
	xmlSetProp(subtree, "amp", buffer);
	sprintf (buffer, "%f", track->pan);
	xmlSetProp(subtree, "pan", buffer);
	sprintf (buffer, "%d", track->mute_flag);
	xmlSetProp(subtree, "mute", buffer);
	g_list_foreach (track->block_list, (GFunc)save_block, subtree);
}

/*
 * Save a mix file.
 */
gboolean
song_save (Song *song, const gchar *filename)
{
	xmlDocPtr doc;
	xmlNsPtr mixmagic_ns;

	doc = xmlNewDoc("1.0");
	doc->root = xmlNewDocNode(doc, NULL, "mixfile", NULL);
	mixmagic_ns = xmlNewNs (doc->root, 
				  "http://mixmagic.sourceforge.net/v1/", "mixmagic");
	xmlSetNs (doc->root, mixmagic_ns);

	g_hash_table_foreach (song->sample_hash, (GHFunc)save_sample, doc->root);

	g_list_foreach (song->track_list, (GFunc)save_track, doc->root);

	if (xmlSaveFile (filename, doc) == -1) {

		g_print ("Failed to save %s\n", filename);
		return FALSE;
	}

	xmlFreeDoc (doc);

	return TRUE;
}

/*
 * Loads a mix file.
 */
gboolean
song_load (Song *song, const gchar *filename)
{
	xmlDocPtr doc;
	xmlNodePtr cur, cur2;
	xmlNsPtr ns;
	Sample *sample;
	gint track_num = 0;
	Track *track;
	gchar *ptr;

	g_assert (filename != NULL);

	doc = xmlParseFile (filename);

	if (doc == NULL) {
		
		return FALSE;
	}
	cur = doc->root;
	if (cur == NULL) {

		xmlFreeDoc(doc);
		return FALSE;
	}
	ns = xmlSearchNsByHref(doc, cur, "http://mixmagic.sourceforge.net/v1/");
	if (ns == NULL) {

		g_error ("This is not a mixmagic mixfile:1\n");

		xmlFreeDoc(doc);
		return FALSE;
	}
	if (strcmp(cur->name, "mixfile")) {

		g_error ("This is not a mixmagic mixfile:2\n");
		xmlFreeDoc(doc);
		return FALSE;
	}

	cur = cur->childs;
	while (cur != NULL) {

		if ((!strcmp(cur->name, "sample")) && (cur->ns == ns)) {
			/* Allocate a new sample stucture */
			sample = sample_new ();

			sample->id = xmlGetProp (cur, "id");
			sample->filename = xmlGetProp (cur, "filename");
			/* Let the sample initialize itself */
			sample_init (sample, song->appbar);
			/* Add the sample to our list */
			g_hash_table_insert (song->sample_hash, sample->id, sample);
		}
		if ((!strcmp(cur->name, "track")) && (cur->ns == ns)) {

			cur2 = cur->childs;
			while (cur2 != NULL) {
				if ((!strcmp(cur2->name, "block")) && (cur2->ns == ns))
					block_load (song, NULL, cur2, ns, track_num);

				if ((!strcmp(cur2->name, "loopblock")) && (cur2->ns == ns))
					loopblock_load (song, NULL, cur2, ns, track_num);

				cur2 = cur2->next;
			}
			track = TRACK (g_list_nth (song->track_list, track_num)->data);
			track_set_amp (track, atof ((ptr = xmlGetProp (cur, "amp"))));
			g_free (ptr);
			track_set_pan (track, atof ((ptr = xmlGetProp (cur, "pan"))));
			g_free (ptr);
			track_set_mute (track, atoi ((ptr = xmlGetProp (cur, "mute"))));
			g_free (ptr);

			track_num++;
		}
		cur = cur->next;
	}
	xmlFreeDoc(doc);

	return TRUE;
}

void
song_delete_block (Song *song, Block *block)
{
	GList *list;
	Track *track;

	list = song->track_list;

	while (list) {
		
		track = TRACK (list->data);
		if (g_list_find (track->block_list, block)) {

			/* Remove the sample if this block is the last to use it */
			if (block->sample->ref_count == (1 + 1)) { /* Block + sample_hash */
				sample_unref (block->sample);
				g_hash_table_remove (song->sample_hash, block->sample->id);
			}
			/* Free the block */
			track_remove_block (track, block);

			/* Did the Track become empty? */
			if (g_list_length (track->block_list) == 0) {

				list = g_list_next (list);

				if (g_list_next (list) == NULL) { /* Is this the last track on the list */
					/* Remove empty tracks from the bottom */
					while (list && 
					       g_list_length (TRACK(list->data)->block_list) == 0 &&
					       g_list_length (song->track_list) > 1 &&
					       g_list_length (TRACK(g_list_previous (list)->data)->block_list) == 0) {

						track = TRACK (list->data);
						gtk_widget_destroy (GTK_WIDGET (track));
						song->track_list = g_list_remove (song->track_list, track);
						list = g_list_previous (list);
					}
				}
			}
			break;
		}
		else
			list = g_list_next (list);
	}
}

/*
 * Bad hack to find an entry in a hash table.
 */
static void
find_sample_by_name (gpointer key, gpointer value, gpointer user_data)
{
	Sample *sample = (Sample *)value;
	gpointer *ptr = (gpointer *)user_data;
	
	if (ptr[1] != NULL)
		return;
	if (strcmp (ptr[0], sample->filename) == 0)
		ptr[1] = sample;
}

/*
 * Return a sample structure for filename, allocates it if needed.
 */
Sample *
song_get_sample (Song *song, gchar *filename) 
{
	Sample *sample;
	gpointer ptr[2], tmp;

	g_assert (song != NULL);
	g_assert (filename != NULL);

	ptr[0] = filename;
	ptr[1] = NULL;
	/* First try to find it in the hash */
	g_hash_table_foreach (song->sample_hash, find_sample_by_name, ptr);
	if (ptr[1])
		return (Sample *)ptr[1];
	
	/* Nope, have to create a new sample */
	sample = sample_new ();
	sample->filename = g_strdup (filename);

	/* Try to initialize it */
	if (sample_init (sample, song->appbar) == FALSE) {

		sample_unref (sample);
		return NULL;
	}
	/* Create an unique id */
	sample->id = g_strdup (g_basename (sample->filename));
	while (g_hash_table_lookup (song->sample_hash, sample->id)) {
		tmp = sample->id;
		sample->id = g_strdup_printf ("%s%s", sample->id, "0");
		g_free (tmp);
	}
	/* Add the sample to the hash */
	g_hash_table_insert (song->sample_hash, sample->id, sample);
	return sample;
}

void
song_insert_block (Song *song, Block *block, gint track_num)
{
	Track *track;
	gint nr_tracks;

	nr_tracks = g_list_length (song->track_list);
	/*Create a new track (+1)if it doesn't exist */
	while (nr_tracks < (track_num + 2)) {

		track = TRACK (track_new (song_get_adjustment (song), song, nr_tracks));
		gtk_widget_show (GTK_WIDGET (track));
		gtk_box_pack_start (GTK_BOX (song),
				    GTK_WIDGET (track), FALSE, FALSE, 0);

		song->track_list = g_list_append (song->track_list, track);

		nr_tracks++;
	}

	/* Add the sample to the hash, if it's missing */
	if (g_hash_table_lookup (song->sample_hash, block->sample->id) == NULL)
		g_hash_table_insert (song->sample_hash, block->sample->id,
				     sample_ref (block->sample));
	
	track = g_list_nth_data (song->track_list, track_num);

	track_add_block (track, block);
}

static void
update_track (gpointer data, gpointer user_data)
{
	track_update (TRACK (data));
}

void
song_update (Song *song)
{
	g_list_foreach (song->track_list, update_track, NULL);
}

static void
update_marker_on_track (gpointer data, gpointer user_data)
{
	track_move_marker (TRACK (data), GPOINTER_TO_INT (user_data));
}

static gint
timeout_update_marker (gpointer data)
{
	Song *song = SONG (data);
	GtkWidget *widget = GTK_WIDGET (song);
	gchar buffer[80];
	gint m, s, ms;

	song_move_marker (song, -1);
	
	if (mixer_is_playing () == FALSE) {
		song_stop_playback (song);

		if (song->appbar)
			gnome_appbar_pop (GNOME_APPBAR (song->appbar));

		return FALSE;
	}

	if (song->marker_position / song->x_scale - widget->allocation.width / 2 >
	    song_get_adjustment (song)->value &&
	    song_get_length (song) / song->x_scale > GTK_WIDGET(song->track_list->data)->allocation.width) {

		gtk_adjustment_set_value (song_get_adjustment (song), 
					  song->marker_position / song->x_scale - widget->allocation.width / 2);

	}
	if (song->marker_position / song->x_scale <
	    song_get_adjustment (song)->value) {

		gtk_adjustment_set_value (song_get_adjustment (song), 
					  song->marker_position / song->x_scale - widget->allocation.width / 2);
	}


	if (song->appbar) {
		ms = song->marker_position * 100 / SAMPLE_RATE;
		
		m = ms / 6000;
		s = ms % 6000 / 100;
		ms = (ms % 6000) % 100;
		sprintf (buffer, "%02d:%02d:%02d", m, s, ms);
		
		gnome_appbar_set_status (GNOME_APPBAR (song->appbar), buffer);
	}
	return TRUE;
}

void
song_play (Song *song)
{
	song->marker_id = gtk_timeout_add (1000/11, timeout_update_marker, song);
	mixer_play_song_threaded (song);
}

void
song_play_block_and_delete (Song *song, Block *block)
{
	song_move_marker (song, block->position);
	song->marker_id = gtk_timeout_add (1000/11, timeout_update_marker, song);
	mixer_play_block_and_delete (song, block);
}

void
song_stop_playback (Song *song)
{
	mixer_stop_thread ();
}

void
song_update_scrollbar (Song *song)
{
	gint length;
	GList *list;
	Track *track;

	g_assert (song != NULL);
	g_assert (IS_SONG (song));

	length = song_get_length (song);

	list = song->track_list;

	while (list) {
		track = TRACK (list->data);

		track_set_length (track, length / song->x_scale);

		list = g_list_next (list);
	}
}

/*
 * This function returns the length of the song in samples.
 */
gint
song_get_length (Song *song) 
{
	GList *ptr, *ptr2;
	Track *track;
	Block *block;
	gint length = 0;

	g_assert (song != NULL);
	g_assert (IS_SONG (song));

	ptr = song->track_list;

	while (ptr) {
		track = TRACK (ptr->data);

		ptr2 = track->block_list;

		while (ptr2) {
			
			block = BLOCK (ptr2->data);
			if (block->position + block->length > length)
				length = block->position + block->length;

			ptr2 = g_list_next (ptr2);
		};

		ptr = g_list_next (ptr);
	};

	return length;
}

/*
 * This function adjusts the scale value to zoom in.
 */
void
song_zoom_in (Song *song) 
{
	g_assert (song != NULL);
	g_assert (IS_SONG (song));

	song->x_scale /= 1.2;

	song_update (song);
}
/*
 * This function adjusts the scale value to zoom out.
 */
void
song_zoom_out (Song *song) 
{
	g_assert (song != NULL);
	g_assert (IS_SONG (song));

	song->x_scale *= 1.2;

	song_update (song);
}
/*
 * This function adjusts the scale value to fit all blocks on the screen.
 */
void
song_zoom_fit (Song *song) 
{
	gint length, width;

	g_assert (song != NULL);
	g_assert (IS_SONG (song));

	length = song_get_length (song);
	width = TRACK (song->track_list->data)->layout->allocation.width;

	song->x_scale = length / width;

	song_update (song);
}

void song_move_marker (Song *song, gint position)
{
	g_assert (song != NULL);
	g_assert (IS_SONG (song));

	if (position != -1)
		song->marker_position = position;

	g_list_foreach (song->track_list, update_marker_on_track, GINT_TO_POINTER(song->marker_position/song->x_scale));
}

